/*-- REXX --------------------------------------------------------------------*/
/*                                                                            */
/* Module:      rspcomp.rex                                                   */
/*                                                                            */
/* Version:     1.1.1                                                         */
/*                                                                            */
/* Author:      W. David Ashley                                               */
/*                                                                            */
/* Description: Compile an *.rsp (REXX Server Page) into a real REXX pgm.     */
/*                                                                            */
/* Copyright (C) International Business Machines Corporation and others.      */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* This software is subject to the terms of the Common Public License. You    */
/* must accept the terms of this license to use this software. Refer to       */
/* the file CPLv1.0.htm included in this package for more information.        */
/*                                                                            */
/* The program is provided "as is" without any warranty express or implied,   */
/* including the warranty of non-infringement and the implied warranties of   */
/* merchantibility and fitness for a particular purpose. IBM will not be      */
/* liable for any damages suffered by you as a result of using the Program.   */
/* In no event will IBM be liable for any special, indirect or consequential  */
/* damages or lost profits even if IBM has been advised of the possibility of */
/* their occurrence. IBM will not be liable for any third party claims        */
/* against you.                                                               */
/*                                                                            */
/* Modifications:                                                             */
/*                                                                            */
/* Date       Author      Description                                         */
/* ---------- ----------- --------------------------------------------------- */
/* 2002/06/01 WD Ashley   v1.1.0 Initial Public Release                       */
/* 2002/06/30 WD Ashley   v1.1.1 Changed the license reference only.          */
/* 2002/09/14 J.M.Blasco  TAG1   Initial simple rexx tag approach             */
/*                                                                            */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Global option variables:                                                   */
/*                                                                            */
/* errmsg   = 0  do not show error messages (default)                         */
/*            1  output error messages to stdout                              */
/*                                                                            */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Global state variable:                                                     */
/*                                                                            */
/* state    = 0  parsing HTML lines (starting state)                          */
/*            1  parsing REXX statements delimited by '<?rexx' and '?>' tags  */
/*            2  parsing REXX statements delimited by                         */
/*               '<script type="rexx">' and '</script>' tags                  */
/*                                or                                          */
/*               '<script language="rexx">' and '</script>' tags              */
/*                                                                            */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Return codes:                                                              */
/*                                                                            */
/* 0  successful execution                                                    */
/* 1  no output REXX filename specified                                       */
/* 2  the input RSP file was not found                                        */
/* 3  cannot open the RSP input file                                          */
/* 4  cannot open the REXX output file                                        */
/*                                                                            */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/* Simple TAG support                                                         */
/*                                                                            */
/* Added by J. M. Blasco <jm@jmblasco.com>, 14/09/2002                        */
/*                                                                            */
/* We implement a very simple form of tag support. We parse HTML lines (i.e., */
/* state = 0), both in start-tag form (<tag [parms]>), end-form (</tag>)      */
/* and auto-ended form <tag [parms]/>. For every tag <tagname>, we call       */
/*   call <tagname>.tag verb [extra], parms                                   */
/* 'verb' can be: START for start tags or END for end tags. Other values are  */
/* reserved for future expansion. 'parms' are the parameters supplied to the  */
/* tag (only for start or auto-ended forms, no parsing).                      */
/*                                                                            */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*                                                                            */
/* When a tag wants to return multiple lines, it does so by separating lines  */
/* with '00' characters. The compiler breaks the result into different lines, */
/* which are manipulated using a lifo stack.                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/

validchars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvxyz0123456789'

linestack.0 = 0 /* The stack is empty */

tagstack.0 = 0 /* To match for end-tags */
parmstack.0 = 0 /* To pass parameters to the end-tag */

tagcache. = 0 /* We don't have cache information by default */

lineno = 0 /* For error msgs */

/* version of this script */
script_version = 'v1.1.1'

/* set global variables and option defaults */
errmsg = 0
parse upper version rexx_ver

/* get the input filename arguments and options */
cmdline = arg(1)
if substr(cmdline, 1, 1) = '"' then ,
 parse var cmdline '"' rspfilename '"' cmdline
else ,
 parse var cmdline rspfilename cmdline
cmdline = strip(cmdline)
if substr(cmdline, 1, 1) = '"' then ,
 parse var cmdline '"' rexfilename '"' option1 .
else ,
 parse var cmdline rexfilename option1 .
if rexfilename = '' then do
   call console_msg 'Error: No output REXX filename specified.'
   return 1
   end
if translate(option1) = 'ERRMSG' then errmsg = 1

/* try to open the rsp file */
if rspfilename = '' then do
   call console_msg 'Error: No input RSP File specified.'
   return 2
   end
retc = stream(rspfilename, 'c', 'open read')
if retc <> 'READY:' then do
   call console_msg 'Error: cannot open file' rspfilename
   return 3
   end

/* open the rex file */
/*retc = stream(rexfilename, 'c', 'open write replace')*/
retc = stream(rexfilename, 'c', 'open write')
if retc <> 'READY:' then do
   call console_msg 'Error: cannot open file' rexfilename
   return 4
   end

/* set up an end-of-file indicator on the input rsp file */
eof = 0
call on notready name seteof

/* read in the rsp file and look for the the rsp tags */
state = 0 /* initial state is to output HTML lines */
call REXX_pgm_header

line = getline()
do while eof = 0
   call process_line
   line = getline()
   end

/* close the files */
call stream rspfilename, 'c', 'close'
call stream rexfilename, 'c', 'close'

/* done */
return 0

/*----------------------------------------------------------------------------*/
/* function which returns the next line (from the input file or the stack)    */
/*----------------------------------------------------------------------------*/

getline: procedure expose linestack. eof rspfilename lineno
   If linestack.0 > 0 then Do
     i = linestack.0
     line = linestack.i
     linestack.0 = i - 1
     End
   Else Do
     line = LineIn(rspfilename)
     if eof = 1 then line = ''
     else lineno = lineno + 1
   End
return line 

/*----------------------------------------------------------------------------*/
/* function which indicates end-of-file                                       */
/*----------------------------------------------------------------------------*/

seteof:
eof = 1
return


/*----------------------------------------------------------------------------*/
/* process a line from the rsp file                                           */
/*----------------------------------------------------------------------------*/

process_line: 
uline = translate(strip(line))
select
   when state = 0 then do
      if uline = '<?REXX' then do
         /* change state to output REXX pgm line(s) */
         state = 1
         end
      else if substr(uline, 1, 8) = '<SCRIPT ' then do
         parse var uline '<SCRIPT' . 'TYPE=' lang . '>'
         if lang = '' then ,
          parse var uline '<SCRIPT' . 'LANGUAGE=' lang . '>'
         lang = strip(lang,, '"')
         lang = strip(lang,, "'")
         if lang = 'REXX' then do
            /* change state to output REXX pgm line(s) */
            state = 2
            end
         else nop /* must be another language (like Javascript) */
         end
	 
      If state = 0 Then Do
      
         /*-------------------------------------------------------------------------------------------------*/
	 /* We now parse the line for xml-like tags (i.e. <[/]prefix:tag[parms][/]>                         */
	 /*-------------------------------------------------------------------------------------------------*/
	 
	 colonpos = 0  /* Parsing start point */
         /* Avoid parsing lines which obviously have no tags */
	 if pos(':',line) > 0 & pos('<',line) > 0 then do parsing = 1 to 1
	 
	   /*-----------------------------------------------------------------------------------------------*/
	   /* The following loop exits when:                                                                */
	   /*   1.- We found something resembling a tag (may come back later if we can't associate it with  */
	   /*       a REXX routine).                                                                        */
	   /*   2.- We definitely do not find a tag (and we then leave the whole parsing phase).            */
	   /*-----------------------------------------------------------------------------------------------*/
	 	 
	   Do forever                                                     /* Parse for tags in a line       */
	     pos = colonpos + 1                                           /* Where to look for a colon      */
	     colonpos = Pos(':',line,pos)                                 /* If we don't find one...        */
	   If colonpos == 0 then leave parsing                            /* ...there's no tag.             */
	     ltpos = LastPos('<',line,colonpos-1)                         /* Now look backwards for a '<'.  */
	     If ltpos == 0 then Iterate                                   /* No '<'? Look for next ':'      */
	     prefix = SubStr(line,ltpos+1,colonpos-ltpos-1)               /* Extract the prefix of the...   */
	     pref = prefix                                                /* tag ('<prefix:tag parms>').    */
	     If Left(prefix,1) = '/' then pref = substr(prefix,2)         /* Skip '/' for end-tags          */
	     If Length(pref) == 0 then Iterate                            /* Prefix must not be empty...    */
	     If Verify(pref,validchars) <> 0 then Iterate                 /* ...and chars must be valid.    */
	     gtpos = Pos('>',line,colonpos+1)                             /* Now look for tag termination.  */
	     If gtpos == 0 then Do                                        /* Maybe in a different line?     */
	       blankpos = Pos(' ',line' ',colonpos+1)                     /* Extract the...                 */
	       tag = SubStr(line,colonpos+1,blankpos-colonpos-1)          /* ...tag name.                   */
	       If Verify(tag,validchars) <> 0 then Iterate                /* Chars must be valid            */
	       Do while eof == 0 & gtpos == 0                             /* We now read from the source... */
	         l = getline()                                            /* ...file until we find a '>'... */
		 If eof == 1 then Leave parsing                           /* ...char OR the file is ended.  */
		 line = line l                                            /* We assemble a new line, ...    */
		 gtpos = Pos('>',line,colonpos+1)                         /* ...and reparse for '>'.        */
	       End                                                        /* Now we might have the tag...   */
	     End                                                          /* ...1 line.                     */
	     Parse Value SubStr(line,colonpos+1,gtpos-colonpos-1) with tag parms /* Parse tag and params.   */
	     If Length(tag) == 0 then Iterate                             /* Tag may not be empty, ...      */
	     If Verify(tag,validchars) <> 0 then Iterate                  /* ...and chars must be valid.    */
	     Leave                                                        /* Now the tag is parsed          */
	   end                                                            
	   
	   /*-----------------------------------------------------------------------------------------------*/
	   /* We extract now the 'before' and 'after' parts of the line.                                    */
	   /*-----------------------------------------------------------------------------------------------*/
	   
	   before = Left(line,ltpos-1)
	   after = SubStr(line,gtpos+1)
	   
	   /*-----------------------------------------------------------------------------------------------*/
	   /* Determine the form of the tag. This will influence how the associated REXX routine is called. */
	   /*-----------------------------------------------------------------------------------------------*/
	   
	   If Left(prefix,1) = '/' then Do                                /* '</prefix:tag[parms]>'         */
             tagtype = 'END'
             prefix = SubStr(prefix,2)
             End
           Else If Length(parms) > 0 & Right(parms,1) = '/' then Do       /* '<prefix:tag parms/>'          */
             tagtype = 'BOTH'
             parms = Left(parms,length(parms)-1)
             End
           Else If length(parms) == 0 & right(tag,1) = '/' then Do        /* '<prefix:tag/>'                */
             tagtype = 'BOTH'
             tag = left(tag,length(tag)-1)
             End
           Else Do                                                        /* '<prefix:tag>'                 */
             tagtype = 'START'
           End
	   
	   /*-----------------------------------------------------------------------------------------------*/
	   /* We have now to find a suitable REXX routine that implements the tag in one of the 'taglibs'.  */
	   /*-----------------------------------------------------------------------------------------------*/
	   
	   rexxtagfile = tagcache.prefix.tag                              /* Look in the tag cache          */
	   If rexxtagfile = '' then                                       /* Tag is unknown. Look for...    */
	     Iterate parsing                                              /* ...more tags.                  */
	   
	   If rexxtagfile = 0 then Do                                     /* We don't have cache info       */ 
	     found = 0
	     Do i = 1 to taglib.0
               taglib = taglib.i
               If Right(taglib,1) = '"' then taglib = Left(taglib,length(taglib)-1) /* Apache puts extra "s */
               rexxtagfile = Stream(taglib||tag'.'prefix'.tag','C','QUERY EXISTS')
               If Length(rexxtagfile) > 0 then Do
	         tagcache.tag.prefix = rexxtagfile                        /* Cache the REXX file            */
	         found = 1
                 Leave
	       End
             End
	     If \found then Do                                            /* Tag is unknown.                */
	       tagcache.tag.prefix = ''                                   /* Cache the fact, ..             */
	       Iterate parsing                                            /* ...and look for more tags.     */
	     End
	   End
	   
	   /*-----------------------------------------------------------------------------------------------*/
	   /* We found the associated REXX routine. Call it as appropriate, and, if necessary, stack tag    */
	   /* and parameters for end-tag processing and matching                                            */
	   /*-----------------------------------------------------------------------------------------------*/
	   
	   Select
	     When tagtype = 'START' then Do
               i = tagstack.0 + 1
               tagstack.i = prefix':'tag
               parmstack.i = Strip(parms)
               tagstack.0 = i
	       Interpret 'Call "'rexxtagfile'" tagtype, strip(parms)'
               line = before||result||after
               End /* tagtype = 'START' */
             When tagtype = 'END' then Do
               i = tagstack.0
               If tagstack.i <> prefix':'tag then Do
                 j = linestack.0 + 1
                 If i == 0 Then linestack.j = "<b>REXXTAG Error at line" lineno": &lt;/"prefix":"tag"&gt; has no matching opening tag.</b>"
		 Else linestack.j = "<b>REXXTAG Error at line" lineno": &lt;/"prefix":"tag"&gt; is no matching end-tag for &lt;"tagstack.i"&gt;.</b>"
                 linestack.0 = j
                 Return
               End
               parms = parmstack.i
	       i = i - 1
               tagstack.0 = i
	       Interpret 'Call "'rexxtagfile'" tagtype, strip(parms)'
               line = before||result||after
               End /* tagtype = 'END' */
             When tagtype = 'BOTH' then Do
               Interpret 'Call "'rexxtagfile'" "START", strip(parms)'
               r1 = result
               Interpret 'Call "'rexxtagfile'" "END", strip(parms)'
               r2 = result
               line = before||r1||r2||after
               End /* tagtype = 'BOTH' */
	   End /* Select */
           
	   /*-----------------------------------------------------------------------------------------------*/
	   /* The REXX tag routines may return more than one line. They do so by separating lines with a    */
	   /* '00'x character. We now have to parse the result, and, if necessary, stack the likes LIFO.    */
	   /*-----------------------------------------------------------------------------------------------*/
           
           Do while pos('00'x,line) > 0
             p = LastPos('00'x,line)
	     l = SubStr(line,p+1)
	     line = SubStr(line,1,p-1)
	     i = linestack.0 + 1
             linestack.i = l
             linestack.0 = i
           End
           i = linestack.0 + 1
           linestack.i = line
           linestack.0 = i
	   
           /*-----------------------------------------------------------------------------------------------*/
	   /* The result line(s) have now been stacked. We return WITHOUT outputing the line, so that the   */
	   /* result can be processed again (i.e., re-parsed).                                              */
	   /* This means that tags MAY, if desired, return output which contain other tags, which will be   */
	   /* processed in turn, etc. This also means that a tag can output REXX code to be executed (by    */
	   /* emitting "<script type='REXX'>", etc, and so may leave some state for following tags to use.  */
	   /* This is a little dirty, but it's very easy to use and may be very powerful. There's obviously */
	   /* the danger of recursion (i.e., a <a:b> tag outputs another <a:b> tag, or indirect recursion), */
	   /* but we found that we could not drop all this power just to care for stupid users or           */
	   /* webmasters :-)                                                                                */
           /* This could be an option in a future release (i.e., use a RexxSetVar).                         */
	   /*-----------------------------------------------------------------------------------------------*/
           Return
           
         End /* parsing */

         /* state has not changed, output the HTML line */
         Call lineout rexfilename, "say '" || add_quotes("'", line) || "'"
         if pos('</HTML>', uline) > 0 then do
            /* doing this separates the mainline code from any REXX     */
            /* subroutines/functions added at the end of the RSP file,  */
            /* that way we do not just fall into them (a REXX error).   */
            call REXX_pgm_footer
            end
         end
      end
   when state = 1 then do
      if uline = '?>' then do
         /* change state to output HTML line(s) */
         state = 0
         end
      else do
         /* output the REXX pgm line */
         call lineout rexfilename, line
         end
      end
   when state = 2 then do
      if uline = '</SCRIPT>' then do
         /* change state to output HTML line(s) */
         state = 0
         end
      else do
         /* output the REXX pgm line */
         call lineout rexfilename, line
         end
      end
   otherwise nop
   end
return


/*----------------------------------------------------------------------------*/
/* add quotes to a string (two for one)                                       */
/*----------------------------------------------------------------------------*/

add_quotes: procedure
quote = arg(1)
line = arg(2)
x = pos(quote, line)
do while x > 0
   line = substr(line, 1, x) || quote || substr(line, x + 1)
   x = pos(quote, line, x + 2)
   end
return line


/*----------------------------------------------------------------------------*/
/* add header information to the REXX program                                 */
/*----------------------------------------------------------------------------*/

REXX_pgm_header:
call lineout rexfilename, '/* Start of REXX RSP header information */'
call lineout rexfilename, '/*'
call lineout rexfilename, ''

call lineout rexfilename, 'RSP Compiler Information:'
call lineout rexfilename, '   Compiler version: ' script_version
call lineout rexfilename, '   Compile timestamp:' date() time()
call lineout rexfilename, '   Operating System: ' os
call lineout rexfilename, '   REXX Version:     ' rexx_ver
call lineout rexfilename, ''

call lineout rexfilename, 'Source RSP File Information:'
fsize = stream(rspfilename, 'c', 'query size')
ftimestamp = stream(rspfilename, 'c', 'query timestamp')
call lineout rexfilename, '   ' || ftimestamp || '  ' || format(fsize, 10) || ,
                          '  ' || stream(rspfilename, 'c', 'query exists')
call lineout rexfilename, ''

call lineout rexfilename, 'Output REXX Program File Information:'
call lineout rexfilename, '   Filename:' stream(rexfilename, 'c', 'query exists')
call lineout rexfilename, ''

call lineout rexfilename, '*/'

call lineout rexfilename, ''
call lineout rexfilename, 'call WWWSendHTTPHeader arg(1), "text/html"'
call lineout rexfilename, ''
call lineout rexfilename, '/* End of REXX RSP header information */'

call lineout rexfilename, ''
return


/*----------------------------------------------------------------------------*/
/* add footer information to the REXX program                                 */
/*----------------------------------------------------------------------------*/

REXX_pgm_footer:
call lineout rexfilename, '/* Start of REXX RSP footer information */'
call lineout rexfilename, 'return 0'
call lineout rexfilename, '/* End of REXX RSP footer information */'
return


/*----------------------------------------------------------------------------*/
/* send message to the console (stdout)                                       */
/*----------------------------------------------------------------------------*/

console_msg: 
if errmsg = 1 then say arg(1)
return

